//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include <windows.h>		// Header File For Windows
#include <iostream> 
#include <math.h>

#include "BIHRandTreeBuilder.h"

#define MAX_LEAF_SIZE 8

inline void BIHRandTreeBuilder::BIHExchange(VertexTriangle* triangles, int a, int b)
{
	VertexTriangle tmp = triangles[a];
	triangles[a] = triangles[b];
	triangles[b] = tmp;
}

inline void BIHRandTreeBuilder::BIHNormalExchange(NormalTriangle* normals, int a, int b)
{
	NormalTriangle tmp = normals[a];
	normals[a] = normals[b];
	normals[b] = tmp;
}

inline bool BIHRandTreeBuilder::LeftSmaller(VertexTriangle* triangles, int i, int axis, float pivotValue, float& currentMaxValue)
{
	float maxValue = triangles[i].maxValue(axis);
	if (maxValue < pivotValue)
	{
		currentMaxValue = max(maxValue, currentMaxValue);
		return true;
	}
	else
	{
		float minValue = triangles[i].minValue(axis);
		float midValue = (maxValue + minValue) / 2.0f;
		if (midValue < pivotValue)
		{
			currentMaxValue = max(maxValue, currentMaxValue);
			return true;
		}
		else
			return false;
	}
	
}

inline bool BIHRandTreeBuilder::RightGreaterOrEqual(VertexTriangle* triangles, int j, int axis, float pivotValue, float& currentMinValue)
{
	float minValue = triangles[j].minValue(axis);
	if (minValue >= pivotValue)
	{
		currentMinValue = min(currentMinValue, minValue);
		return true;
	}
	else
	{
		float maxValue = triangles[j].maxValue(axis);
		float midValue = (maxValue + minValue) / 2.0f;
		if (pivotValue <= midValue)
		{
			currentMinValue = min(currentMinValue, minValue);
			return true;
		}
		else
			return false;
	}
}

int BIHRandTreeBuilder::BIHPartition(VertexTriangle* triangles, NormalTriangle* normals, int left, int right, float pivotValue, int axis, float& maxValue, float& minValue)
{
	int i = left;
	int j = right;

	maxValue = -FLT_MAX;
	minValue = FLT_MAX;

	while (true)
	{
		while (LeftSmaller(triangles, i, axis, pivotValue, maxValue))
		{	
			i++;
		}
		while (j > left && RightGreaterOrEqual(triangles, j, axis, pivotValue, minValue))
		{
			j--;	
		}
		if (i >= j)
		{
			break;
		}
		else
		{
			BIHExchange(triangles, i, j);
			BIHNormalExchange(normals, i, j);
			maxValue = max(triangles[i].maxValue(axis), maxValue);
 			minValue = min(triangles[j].minValue(axis), minValue);
			i++;
			j--;
		}
	}
	if (j == left) //no max value for left side
		maxValue = triangles[left].maxValue(axis);
	return j;
}

void BIHRandTreeBuilder::BIHSort(VertexTriangle* triangles, NormalTriangle* normals, int left, int right, int axis, float& maxValue, float& minValue)
{
	if (right == left)
		return;
	else
	{
		int randomIndex = (rand() % (left - right)) + left;
		float seperationAxis = triangles[randomIndex].midValue(axis);
		int leftIndex = BIHPartition(triangles, normals, left, right, seperationAxis, axis, maxValue, minValue);
		float subMax;
		float subMin;
		BIHSort(triangles, normals, left, leftIndex, axis, subMax, subMin);
		BIHSort(triangles, normals, leftIndex+1, right, axis, subMax, subMin);
	}
}

void BIHRandTreeBuilder::BIHBuildTree(BIHNode* & node, VertexTriangle* triangles, NormalTriangle* normals, int start, int end, Vector min, Vector max)
{
	int count = end - start + 1;
	if (count <= MAX_LEAF_SIZE) // leaf node
	{
		node = new BIHNode((unsigned int)start, count);
	}
	else
	{
		Vector length = max - min;

		int maxLengthAxis = length.maxValueAxis();
		int randomIndex = (rand() % (end - start)) + start;
		int randomIndex2 = (rand() % (end - start)) + start;
		int randomIndex3 = (rand() % (end - start)) + start;
		float seperationAxis1 = triangles[randomIndex].midValue(maxLengthAxis);
		float seperationAxis2 = triangles[randomIndex2].midValue(maxLengthAxis);
		float seperationAxis3 = triangles[randomIndex3].midValue(maxLengthAxis);
		
		float seperationAxis = max(seperationAxis1, min(seperationAxis3, seperationAxis2));
		
		node = new BIHNode((unsigned char)maxLengthAxis);
		// cool, now do partition
		float leftBoundary;
		float rightBoundary;
		/*BIHSort(triangles, normals, start, end, maxLengthAxis, leftBoundary, rightBoundary);
		int leftIndex = (end - start)/2 + start;*/
		int leftIndex = BIHPartition(triangles, normals, start, end, seperationAxis, maxLengthAxis, leftBoundary, rightBoundary);
		
		node->inner.boundaries[0] = leftBoundary;
		node->inner.boundaries[1] = rightBoundary;
		
		BIHBuildTree(node->inner.childs[0], triangles, normals, start, leftIndex, min, Vector(max, maxLengthAxis, leftBoundary));
		BIHBuildTree(node->inner.childs[1], triangles, normals, leftIndex+1, end, Vector(min, maxLengthAxis, rightBoundary), max);
	}
}


BIHTree* BIHRandTreeBuilder::Compile(Scene* scene)
{
	VertexTriangle* triangles = (VertexTriangle*)scene->points;
	NormalTriangle* normals = (NormalTriangle*)scene->normals;
	int triangleCount = scene->count / 3;
	VertexTriangle* start = triangles;
	Vector min;
	Vector max;
	Helpers::GetBoundingBox(scene->points, scene->count, min, max);
	BIHTree* tree = new BIHTree(min, max);
	BIHBuildTree(tree->root, triangles, normals, 0, triangleCount - 1, min, max);
	assert(triangleCount == Helpers::CountItems(tree->root));
	return tree;
}





