//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include <vector>
#include <queue>
#include <math.h>
#include <algorithm> //for sort
#include <float.h>

#include "BIHSAHTreeBuilder.h"

#define TRAVCOST 1.0F
#define INTRCOST 2 * TRAVCOST

#define MAX_LEAF_SIZE 8

float RecursionCosts(float left_size, int left_count, float right_size, int right_count, float parent_size)
{
	return TRAVCOST + INTRCOST * (left_size * left_count + right_size * right_count) / parent_size;
}

float IntersectionCosts(int count)
{
	return count * INTRCOST;
}

void BIHSAHTreeBuilder::BIHExchange(Event* events, int a, int b)
{
	Event tmp = events[a];
	events[a] = events[b];
	events[b] = tmp;
}

int BIHSAHTreeBuilder::BIHPartition(Event* events, int left, int right, float pivotValue, float* primitive_maxs)
{
	int i = left;
	int j = right;

	while (true)
	{
		while (i < right && primitive_maxs[events[i].primitive] <= pivotValue)
		{	
			i++;
		}
		
		while (j > left && primitive_maxs[events[j].primitive] > pivotValue)
		{
			j--;	
		}
		if (i >= j)
		{
			break;
		}
		else
		{
			BIHExchange(events, i, j);
			assert(primitive_maxs[events[i].primitive] <= pivotValue);
			assert(primitive_maxs[events[j].primitive] > pivotValue);
			i++;
			j--;
		}
	}
	return j;
}


BIHNode* BIHSAHTreeBuilder::BuildTreeRecursivly(Event* events[3], float* primitive_mins[3], float* primitive_maxs[3], int event_start, int event_end, Vector boundary_min, Vector boundary_max, int* result_indices)
{

	int eventCount = event_end - event_start + 1;
	assert((eventCount % 2) == 0);
	int triangleCount = eventCount / 2;
	//assert(triangleCount > 0);
	if (triangleCount > 0)
	{
		if (triangleCount > MAX_LEAF_SIZE)
		{
			// build helper arrays to get the right_size for any splitting candidate
			float* right_mins[3] = { new float[triangleCount], new float[triangleCount], new float[triangleCount]};
			for (int axis = 0; axis < 3; axis++)
			{
				int position = triangleCount-1;
				float right_min = boundary_max[axis];
				
				for (int index = event_end; index >= event_start; index--)
				{
					Event e = events[axis][index];
					if (e.type == Type::End)
					{
						right_mins[axis][position] = right_min;
						right_min = min(right_min, primitive_mins[axis][e.primitive]);
						position--;
					}
					//right_mins[axis][position] = right_min;
				}
				int a = 0;
			}

			// find best split candidate
			float current_cost = FLT_MAX;
			int current_index;
			int current_axis;
			float current_left_max;
			float current_right_min;
			float current_position;

			float intersection_cost = IntersectionCosts(triangleCount);

			for (int axis = 0; axis < 3; axis++)
			{
				float min = boundary_min[axis];
				float max = boundary_max[axis];
				int left_count = 0;
				float left_max = boundary_min[axis];
				float right_min = boundary_min[axis];
				for (int index = event_start; index <= event_end; index++)
				{
					Event e = events[axis][index];
					float position = e.position(primitive_mins[axis], primitive_maxs[axis]);
					while(true)
					{
						if (e.type == Type::End)
						{
							right_min = right_mins[axis][left_count];
							left_count++;
							assert(position >= left_max);
							left_max = position;
						}
						if (index+1 > event_end)
							break;
						else
						{
							Event new_event = events[axis][index+1]; 
							float new_position = new_event.position(primitive_mins[axis], primitive_maxs[axis]);
							if (position != new_position)
								break;
							else
							{
								index++;
								e = new_event;
							}
						}	
					}

					int right_count = triangleCount - left_count;

					float candidate_cost = 
						RecursionCosts(left_max - min, left_count, max - right_min, right_count, max - min);
					if (candidate_cost < current_cost)
					{
						current_cost = candidate_cost;
						current_index = index;
						current_axis = axis;
						current_left_max = left_max;
						current_right_min = right_min;
						current_position = position;
					}
				}
			}
			
			// check wether cost are below intersection costs
			if ( current_cost < intersection_cost)
			{
				// first partition events
				
				int left_end;
				int prev_left_end = -1;
				for (int axis = 0; axis < 3; axis++)
				{
					left_end = BIHPartition(events[axis], event_start, event_end, current_position, primitive_maxs[current_axis]);
					if (prev_left_end >= 0)
						assert(left_end == prev_left_end);
					prev_left_end = left_end;
					// in place partition is unstable, so we have to sort again :(
					sort(events[axis]+event_start,events[axis]+left_end+1, EventCompare(primitive_mins, primitive_maxs, axis));
					sort(events[axis]+left_end+1,events[axis]+event_end+1, EventCompare(primitive_mins, primitive_maxs, axis));

					if ( (left_end - event_start + 1) > 0)
					{
						float mymax = -FLT_MAX;
						int max_index = -1;
						for (int test = event_start; test <= left_end; test++)
						{
							float p = events[axis][test].position(primitive_mins[current_axis], primitive_maxs[current_axis]);
							if (p > mymax)
							{
								max_index = test;
								mymax = p;
							}
							assert(p <= current_position);
						}
						assert(mymax == current_left_max);
					}
					if ((event_end - (left_end+1) +1) > 0)
					{
						int min_index = -1;
						float mymin = FLT_MAX;
						for (int test = left_end+1; test <= event_end; test++)
						{
							float p = events[axis][test].position(primitive_mins[current_axis], primitive_maxs[current_axis]);
							if (events[axis][test].type == Type::End)
							{
								assert(p > current_position);
							}
							if (p < mymin)
							{
								mymin = p;
								min_index = test;
							}
						}
						assert(mymin == current_right_min);
					}	
				}
				// creating inner node
				BIHNode* inner = new BIHNode(current_axis);
				
				inner->inner.boundaries[0] = current_left_max;
				inner->inner.boundaries[1] = current_right_min;
				// do recursion
				inner->inner.childs[0] = BuildTreeRecursivly(events, primitive_mins, primitive_maxs, event_start, left_end, boundary_min, Vector(boundary_max, current_axis, current_left_max), result_indices);
				inner->inner.childs[1] = BuildTreeRecursivly(events, primitive_mins, primitive_maxs, left_end+1, event_end, Vector(boundary_min, current_axis, current_right_min), boundary_max, result_indices);
				return inner;
			}

		}
		
		// Create Leaf Node

		int index = event_start/2;
		for (int e = event_start; e <= event_end; e++)
		{
			Event ev = events[0][e];
			if (ev.type == Type::End)
			{
				result_indices[index] = ev.primitive;
				index++;
			}
		}
			// create leaf node
		return new BIHNode(event_start/2, triangleCount);
	}
	else
	{
		// Create Empty Leaf Node
		return new BIHNode(-1, 0);
	}
		
}

BIHTree* BIHSAHTreeBuilder::Compile(Scene* scene)
{
	VertexTriangle* triangles = (VertexTriangle*)scene->points;
	NormalTriangle* normals = (NormalTriangle*)scene->normals;
	
	int triangleCount = scene->count / 3;
	int* sortedIndices = new int[triangleCount];
	VertexTriangle* start = triangles;
	Vector min;
	Vector max;
	Helpers::GetBoundingBox(scene->points, scene->count, min, max);
	float* mins[3] = { new float[triangleCount], new float[triangleCount], new float[triangleCount]};
	float* maxs[3] = { new float[triangleCount], new float[triangleCount], new float[triangleCount]};
    Event* events[3] = {new Event[triangleCount*2], new Event[triangleCount*2], new Event[triangleCount*2]};

	for (int axis = 0; axis < 3; axis++)
	{
		for (int index = 0; index < triangleCount; index++)
		{
			events[axis][2*index] = Event(index, Type::Start);
			events[axis][2*index+1] = Event(index, Type::End);
			VertexTriangle t = triangles[index];
			mins[axis][index] = t.minValue(axis);
			maxs[axis][index] = t.maxValue(axis);
		}
	}
	for (int axis = 0; axis < 3; axis++)
	{
		sort(events[axis], events[axis]+triangleCount*2, EventCompare(mins, maxs, axis));
		for (int test = 0; test < triangleCount*2-1; test++)
		{
			float p1 = events[axis][test].position(mins[axis], maxs[axis]);
			float p2 = events[axis][test+1].position(mins[axis], maxs[axis]);
			assert(p1 <= p2);
		}
	}

	int* indices = new int[triangleCount];

	BIHNode* root = BuildTreeRecursivly(events, mins, maxs, 0, triangleCount*2-1, min, max, indices);
	VertexTriangle* new_triangles = new VertexTriangle[triangleCount];
	NormalTriangle* new_normals = new NormalTriangle[triangleCount];
	for (int index = 0; index < triangleCount; index++)
	{
		int source_index = indices[index];
		new_triangles[index] = triangles[source_index];
		new_normals[index] = normals[source_index];
	}
	scene->points = (Point*)new_triangles;
	scene->normals = (Normal*)new_normals;

	delete triangles;
	delete normals;
	delete indices;
	for (int axis = 0; axis < 3; axis++)
	{
		delete events[axis];
		delete mins[axis];
		delete maxs[axis];
	}
	
	BIHTree* tree = new BIHTree(min,max);
	tree->root = root;
	
	return tree;
}