//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include <iostream>  
#include <math.h>
#include <assert.h>
#include <float.h>

#include "Raytracer.h"
#include "Sampler.h"
#include "Settings.h"


#define MAXI 1
#define MINI 0

Intersection::Intersection(float t, unsigned int index, float beta, float gamma): t(t), index(index), beta(beta), gamma(gamma)
{
}
	
bool Intersection::operator==(const Intersection& i)
{
	return t==i.t && index==i.index;
}

Intersection Intersection::NO_INTERSECTION = Intersection(FLT_MAX,0,0,0);

bool Intersection::operator<(const Intersection& i)
{
	return t < i.t;
}

bool Intersection::operator>(const Intersection& i)
{
	return t > i.t;
}

Intersection::Intersection(const Intersection& i): t(i.t), index(i.index), beta(i.beta), gamma(i.gamma)
{
}

bool Raytracer::Intersect(const Ray & r, const Vector bounds[2], float tmin, float tmax, float r_interval[2])
{
	float n_tmin, n_tmax;
	for (unsigned int index = 0; index < 3; index++)
	{
		n_tmin = (bounds[r.sign[index]][index] - r.origin[index]) * r.inv_direction[index];
		n_tmax = (bounds[1-r.sign[index]][index] - r.origin[index]) * r.inv_direction[index];
		if (n_tmin > tmax || n_tmax < tmin)
			return false;
		if (n_tmin > tmin)
			tmin = n_tmin;
		if (n_tmax < tmax)
			tmax = n_tmax;
	}
	r_interval[0] = tmin;
	r_interval[1] = tmax;
	return true;
}


bool Raytracer::Intersect(const Ray& r, const TriAccel& acc, float& t, float& beta, float& gamma, float tMin, float tMax)
{
	int ku = modulo[acc.k+1];
	int kv = modulo[acc.k+2];
	const float nd = 1./
		(r.direction[acc.k] + acc.n_u * r.direction[ku] + acc.n_v*r.direction[kv]);
	t = (acc.n_d - r.origin[acc.k] - acc.n_u * r.origin[ku] - acc.n_v*r.origin[kv])*nd;
	if (t > tMax || t < tMin)
		return false;
	const float hu = r.origin[ku] + t * r.direction[ku];
	const float hv = r.origin[kv] + t * r.direction[kv];
	beta = (hu*acc.b_nu+hv*acc.b_nv+acc.b_d);
	if (beta < 0.0f)
		return false;
	gamma = (hu*acc.c_nu+hv*acc.c_nv+acc.c_d);
	if (gamma < 0.0f)
		return false;
	if (beta+gamma > 1.0f)
		return false;
	return true;
}

bool Raytracer::Intersect(const Ray & r, const Point & A, const Point & B, const Point & C, const Normal & n, float& t, float& beta, float& gamma)
{
	float div = r.direction*n;
	if ( div == 0 )
		return false;
	else {
		t = -((r.origin-A)*n)/div;
		if ( t < MIN_DISTANCE || t > MAX_DISTANCE )
			return false;
		else {
			int k;
			if ( fabs(n.x) > fabs(n.y) )
				if ( fabs(n.x) > fabs(n.z) )
					k = 0;
				else
					k = 2;
			else
				if ( fabs(n.y) > fabs(n.z) )
					k = 1;
				else
					k = 2;
			int u = modulo[k+1];
			int v = modulo[k+2];

			Point hit = r*t;

			Point b = A-B;
			Point c = C-B;
			Point h = hit-B;

			// (b[u] * c[v] - b[v] * c[u])
			float div2 = b[u]*c[v]-b[v]*c[u];
			if ( div2 == 0 )
				return false;
			float div2_inv = 1/div2;
			// beta = (b[u] * H[v] - b[v] * H[u]) / (b[u] * c[v] - b[v] * c[u]);
			beta = (b[u]*h[v] - b[v]*h[u])*div2_inv;
			if ( beta < 0 )
				return false;
			// gamma = (c[v] * H[u] - c[u] * H[v]) / (b[u] * c[v] - b[v] * c[u]);
			gamma = (h[u]*c[v] - h[v]*c[u])*div2_inv;
			if ( gamma < 0 )
				return false;
			if ( (beta + gamma) > 1 )
				return false;
			else {
				return true;
			}
		}
	}
}

Color Raytracer::RaytraceLight(Scene *scene, BIHTree2* tree, const Light& light, const Vector& surfacePosition, const Normal& surfaceNormal)
{
	if (light.type == DirectionalLight)
	{
		if (Settings::ShadowTracing == true)
		{
				Ray ray(surfacePosition, -light.position.normalize());
				if (BoolRaytrace(scene, tree, ray))
					return Color(0,0,0);
				else
					return light.color;
		}
		else
		{
			return light.color;
		}
	}
	else if (light.type == AmbientLight)
	{
		return light.color;
	}
}

bool Raytracer::BoolRaytrace(Scene* scene, BIHTree2* tree, const Ray& ray)
{
	float interval[2];
	Vector bounds[2] = {tree->min, tree->max};
	float tmin = MIN_DISTANCE;
	float tmax = MAX_DISTANCE;
	if (Intersect(ray, bounds, tmin, tmax, interval))
	{
		return BoolRaytrace(tree->root, ray, scene->triAccels, interval[MINI], interval[MAXI]);
	}
	else
		return false;
}

bool Raytracer::BoolRaytrace(const BIHNode2* node, const Ray& ray, TriAccel* triangles, const float tmin, const float tmax)
{
	if (ISLEAF(node->index))
	{
		unsigned int item = INDEX(node->index);
		for (unsigned int index = item; index < item+node->data.leaf.count; index++)
		{
			TriAccel& tri = triangles[index];
			
			float t;
			float beta;
			float gamma;
			if (Intersect(ray, tri, t, beta, gamma, tmin, tmax))
			{
				return true;
			}
		}
		return false;
	}
	else
	{
		unsigned int index = INDEX(node->index);
		unsigned int axis = AXIS(node->index);
		unsigned int sign = ray.sign[axis];

		float t_nearSplit = (node->data.inner.boundaries[sign]-ray.origin[axis]) * ray.inv_direction[axis];
		float t_farSplit = (node->data.inner.boundaries[1-sign]-ray.origin[axis]) * ray.inv_direction[axis];

		if (tmin > t_nearSplit)
			return BoolRaytrace(node+index+1-sign, ray, triangles, max(tmin, t_farSplit), tmax);
		else if (tmax  < t_farSplit)
			return BoolRaytrace(node+index+sign, ray, triangles, tmin, min(tmax, t_nearSplit));
		else
		{
			return BoolRaytrace(node+index+sign, ray, triangles, tmin, min(tmax, t_nearSplit)) || 
				BoolRaytrace(node+index+1-sign, ray, triangles, max(tmin, t_farSplit), tmax);
		}
	}
}

void Raytracer::RaytraceRecursivly(const BIHNode2 * node, const Ray & ray, TriAccel * triangles, const float tmin, const float tmax, Intersection& result)
{
	if (ISLEAF(node->index))
	{
		unsigned int item = INDEX(node->index);
		Intersection intersection = Intersection::NO_INTERSECTION;
		for (unsigned int index = item; index < item+node->data.leaf.count; index++)
		{
			TriAccel& tri = triangles[index];
			
			float t;
			float beta;
			float gamma;
			if (Intersect(ray, tri, t, beta, gamma, tmin, tmax))
			{
				if (t < intersection.t)
				{
					intersection = Intersection(t, index, beta, gamma);
				}
			}
		}
		result = intersection;
	}
	else
	{
		unsigned int index = INDEX(node->index);
		unsigned int axis = AXIS(node->index);
		unsigned int sign = ray.sign[axis];

		float t_nearSplit = (node->data.inner.boundaries[sign]-ray.origin[axis]) * ray.inv_direction[axis];
		float t_farSplit = (node->data.inner.boundaries[1-sign]-ray.origin[axis]) * ray.inv_direction[axis];

		if (tmin > t_nearSplit)
			return RaytraceRecursivly(node+index+1-sign, ray, triangles, max(tmin, t_farSplit), tmax, result);
		else if (tmax  < t_farSplit)
			return RaytraceRecursivly(node+index+sign, ray, triangles, tmin, min(tmax, t_nearSplit), result);
		else
		{
			RaytraceRecursivly(node+index+sign, ray, triangles, tmin, min(tmax, t_nearSplit), result);
			//Raytrace(node+index+sign, ray, triangles, tmin, t_nearSplit, result); //error version
			if (!(result == Intersection::NO_INTERSECTION))
			{
				if  (result.t < t_farSplit)
					return;
				else
				{	
					Intersection i2 = Intersection::NO_INTERSECTION;
					RaytraceRecursivly(node+index+1-sign, ray, triangles, max(tmin, t_farSplit), min(tmax, result.t), i2);
					//Raytrace(node+index+1-sign, ray, triangles, t_farSplit, min(tmax, result.t), i2); //error version
					if (result.t < i2.t)
						return;
					else
					{
						result = i2;
						return;
					}
				}
			}
			else
			{
				RaytraceRecursivly(node+index+1-sign, ray, triangles, max(tmin, t_farSplit), tmax, result);
				//Raytrace(node+index+1-sign, ray, triangles, t_farSplit, tmax, result); //wrong version
				return;
			}
		}
	}
}

float clamp(float f, float max_value)
{
	return min(max_value, f);
}

Color Raytracer::Shade(const Intersection& i, Scene* scene, BIHTree2* tree, const Ray& ray, Normal& normal)
{
	NormalTriangle* nTriangle = ((NormalTriangle*)scene->normals)+i.index;
	if (Settings::LightingModel == Settings::LightingModelType::Flat)
	{
		normal = (nTriangle->a+nTriangle->b+nTriangle->c).normalize();
	}
	else 
	{
		normal = ((nTriangle->b * (1-i.beta-i.gamma)) + (nTriangle->c * i.beta) + (nTriangle->a * i.gamma)).normalize();
	}
	
	Color color(0,0,0);
	Vector hit = ray * i.t;
	for (vector<Light>::iterator it = scene->lights.begin(); it != scene->lights.end(); ++it) {
		Color contribution;
		if (it->HasDirection())
		{
			if (Settings::LightingModel == Settings::LightingModelType::Phong)
			{
				Vector directionToLight = -(it->position).normalize();
				float diff = directionToLight * normal;
				
				if (diff <= 0)
				{
					contribution = Color(0,0,0);
					
				}
				else
				{
					Normal direction_viewer = -ray.direction.normalize();
					float spec = pow(max(normal*(directionToLight+direction_viewer).normalize(),0.0F), 64);
					Color light_color = RaytraceLight(scene, tree, (*it), hit, normal);	
					contribution = light_color.mul(Color(1,1,1)) * (spec+diff);
				}
			}
			else
			{
				Vector directionToLight = -(it->position).normalize();
				float f = directionToLight * normal;
				
				if (f <= 0)
				{
					contribution = Color(0,0,0);
				}
				else
				{
					Color light_color = RaytraceLight(scene, tree, (*it), hit, normal);	
					contribution = light_color.mul( Color(1,1,1)) * f;
				}
			}
		}
		else
		{
			Color light_color = RaytraceLight(scene, tree, (*it), hit, normal);	
			contribution = light_color.mul( Color(1,1,1));
		}
		color = color + contribution;
	}
	return color*(1.0F/scene->lights.size());
	
}

bool Raytracer::Raytrace(Scene* scene, BIHTree2* tree, const Ray& ray, Intersection& i)
{
	float interval[2];
	Vector bounds[2] = {tree->min, tree->max};
	if (Intersect(ray, bounds, MIN_DISTANCE, MAX_DISTANCE, interval))
	{
		RaytraceRecursivly(tree->root, ray, scene->triAccels, interval[MINI], interval[MAXI], i);
		return !(i == Intersection::NO_INTERSECTION);
	}
	else
	{
		return false;
	}
}

void Raytracer::Raytrace(Scene *scene, BIHTree2* tree, const Transform & transform, GLubyte *image, GLsizei imageWidth, GLsizei imageHeight, GLsizei pixel_size)
{
	
	NormalTriangle* normals = (NormalTriangle*)scene->normals;

	const float hw = ((float)imageWidth)/2;
	const float hh = ((float)imageHeight)/2;
	
	const int vertical_block_count = ceil((float)imageHeight / BLOCK_SIZE);
	const int horizontal_block_count = ceil((float)imageWidth / BLOCK_SIZE);

	for ( int block_y = 0 ; block_y < vertical_block_count ; block_y++ ) {
		#pragma omp parallel for
		for ( int block_x = 0 ; block_x < horizontal_block_count ; block_x++ ) {
			int o_x = block_x*BLOCK_SIZE;
			int o_y = block_y*BLOCK_SIZE;

			int vertical_offset = o_y*imageWidth*pixel_size;
			for ( int y = o_y; y < min(BLOCK_SIZE+o_y,imageHeight); y++ ) {

				int horizontal_offset = o_x*sizeof(GLubyte)*4;
				for ( int x = o_x; x < min(BLOCK_SIZE+o_x,imageWidth); x++ ) {
			
					float dx = (float)x - hw;
					float dy = (float)y - hh;
					Normal n(dx/hw, dy/hw, -1);
					Ray ray(transform.position, transform.transformDirection(n));

					GLubyte red = 0;
					GLubyte green = 0;
					GLubyte blue = 0x77;

					Intersection i = Intersection::NO_INTERSECTION;
					if (Raytrace(scene, tree, ray, i))					
					{
						Normal normal;
						Color directLight = Shade(i, scene, tree, ray, normal);
						
						Color incidentLight;
						if (Settings::PathTracing)
						{
							Vector hit = ray * i.t;
							Color indirectLight(0,0,0);
							int count = Settings::PathTracingSamples;
							for (int index = 0; index < count; index++)
							{
								Vector randomDirection = Sampler::GetRandomHemiSpherePoint(normal);
								Ray indirectRay(hit+randomDirection, randomDirection);
								Intersection s = Intersection::NO_INTERSECTION;
								if (Raytrace(scene, tree, indirectRay, s))
								{
									Normal n;
									indirectLight = indirectLight+Shade(s, scene, tree, indirectRay, n);
								}
							}
							incidentLight = indirectLight*(1.0f/count) + directLight;
						}
						else
						{
							incidentLight = directLight;
						}
						
						//Color incidentLight = indirectLight*(1.0f/count);
						
						red = (GLubyte)(min(incidentLight.x*255,255.0f));
						green = (GLubyte)(min(incidentLight.y*255,255.0f));
						blue = (GLubyte)(min(incidentLight.z*255,255.0f));
					}
					int offset = vertical_offset + horizontal_offset;
					image[offset] = red;
					image[offset+sizeof(GLubyte)] = green;
					image[offset+2*sizeof(GLubyte)] = blue;
					image[offset+3*sizeof(GLubyte)] = 0xff;
					horizontal_offset += pixel_size;
				}
				vertical_offset += imageWidth*pixel_size;
			}
		}
	}
}


