//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include "Types.h"
#include "math.h"
#include "float.h"

float Vector::operator *(const Vector & p)
{
	return x*p.x + y*p.y + z*p.z;
}

Vector Vector::operator -()
{
	return Vector(-x,-y,-z);
}

Vector Vector::operator +(const Vector & p)
{
	return Vector(x+p.x, y+p.y, z+p.z);
}

Vector Vector::operator -(const Vector & p)
{
	return Vector(x-p.x, y-p.y, z-p.z);
}

Vector::Vector(const float x, const float y, const float z):x(x), y(y), z(z)
{
}

Vector::Vector(): x(0), y(0), z(0)
{
}

/*Vector& Vector::operator=(const Vector& v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	return *this;
}*/

int Vector::maxValueAxis() const
{
	int maxIndex = -1;
	float maxValue = -FLT_MAX;
	for (int index = 0; index < 3; index++)
	{
		if ((*this)[index] > maxValue)
		{
			maxIndex = index;
			maxValue = (*this)[index];
		}
	}
	return maxIndex;
}

Vector Vector::cross(const Vector & p)
{
	return Vector(y*p.z-z*p.y, z*p.x-x*p.z, x*p.y-y*p.x);
}

Vector Vector::operator *(const float s)
{
	return Vector(x*s, y*s, z*s);
}

inline float& Vector::operator[](const unsigned int i)
{
	return data[i];
}

float Vector::operator*(const Vector& p) const
{
	return x*p.x + y*p.y + z*p.z;
}

Vector Vector::operator+(const Vector& p) const
{
	return Vector(x+p.x, y+p.y, z+p.z);
}

Vector Vector::operator-(const Vector& p) const
{
	return Vector(x-p.x, y-p.y, z-p.z);
}

Vector Vector::operator-() const
{
	return Vector(-x, -y, -z);
}

inline float Vector::operator[](const unsigned int i) const
{
	return data[i];
}

Vector Vector::operator*(const float s) const
{
	return Vector(x*s, y*s, z*s);
}

Vector Vector::normalize() const
{
	float scale = 1/length();
	return *this * scale;
}

float Vector::length() const
{
	return sqrt(x*x+y*y+z*z);
}

float Vector::lengthSquared() const
{
	return x*x+y*y+z*z;
}


Vector::Vector(const Vector & v): x(v.x), y(v.y), z(v.z)
{
}

Vector::Vector(const Vector& v, const int newIndex, const float value)
{
	for (int index = 0; index < 3; index++)
	{
		if (index == newIndex)
			(*this)[index] = value;
		else
			(*this)[index] = v[index];
	}
}

const Vector Vector::UNIT_VECTORS[3] = { Vector(1,0,0), Vector(0,1,0), Vector(0,0,1)};
	
/*float Vector::length()
{
	return sqrt(x*x+y*y+z*z);
}*/

/*float Vector::lengthSquared()
{
	return x*x+y*y+z*z;
}*/

/*Vector Vector::normalize()
{
	float scale = 1/length();
	return *this * scale;
}*/

Vector Vector::mul(const Vector& v)
{
	return Vector(x*v.x, y*v.y, z*v.z);
}

Vector Vector::mul(const Vector& v) const
{
	return Vector(x*v.x, y*v.y, z*v.z);
}

bool Vector::operator==(const Vector& p)
{
	return x == p.x && y == p.y && z == p.z;
}

Vector Vector::rotateX(float angle)
{
	float cosA = cos(angle);
	float sinA = sin(angle);
	return Vector(x, y*cosA-z*sinA, y*sinA+z*cosA);
}

Vector Vector::rotateY(float angle)
{
	float cosA = cos(angle);
	float sinA = sin(angle);
	return Vector(x*cosA+z*sinA, y, -x*sinA+z*cosA);
}

Vector Vector::rotateZ(float angle)
{
	float cosA = cos(angle);
	float sinA = sin(angle);
	return Vector(x*cosA+y*-sinA, x*sinA+y*sinA, z);
}


float VertexTriangle::minValue(unsigned int axis)
{
	return min(min(a[axis], b[axis]), c[axis]);
}

float VertexTriangle::maxValue(unsigned int axis)
{
	return max(max(a[axis], b[axis]), c[axis]);
}

float VertexTriangle::midValue(unsigned int axis)
{
	return (maxValue(axis) + minValue(axis)) / 2.0f;
}

Point Ray::operator*(float t)
{
	return origin+(direction*t);
}

Point Ray::operator*(float t) const
{
	return origin+(direction*t);
}

Ray::Ray(Point o, Normal d):origin(o), direction(d), inv_direction(Vector(1/d.x,1/d.y,1/d.z))
{
	sign[0] = inv_direction.x < 0;
	sign[1] = inv_direction.y < 0;
	sign[2] = inv_direction.z < 0;
}


Light::Light(Vector position, LightType type, Color color): position(position), type(type), color(color)
{
}

bool Light::HasDirection()
{
	return type != LightType::AmbientLight;
}

Scene::~Scene()
{
	delete [] this->normals;
	delete [] this->points;
	delete [] this->triAccels;
}

Scene::Scene()
{
}

BIHTree2::BIHTree2(const BIHNode2* root, const Vector& min, const Vector& max, const int depth, const int count): root(root), min(min), max(max), depth(depth), count(count)
{
}

BIHTree2::~BIHTree2()
{
	delete this->root;
}

BIHNode::BIHNode(unsigned int item, unsigned int count): isLeaf(true)
{
	leaf.item = item;
	leaf.count = count;
}

BIHNode::BIHNode(unsigned char axis): isLeaf(false)
{
	inner.axis = axis;
}

BIHNode::~BIHNode()
{
	if (!this->isLeaf)
	{
		delete this->inner.childs[0];
		delete this->inner.childs[1];
	}
}

BIHNode2::BIHNode2(unsigned int startIndex, unsigned int count): index(startIndex << 2 | 3)
{
	data.leaf.count = count;
}

BIHNode2::BIHNode2(unsigned char axis, unsigned int childrenIndex, float boundaries[2]): index(axis | childrenIndex << 2)
{
	data.inner.boundaries[0] = boundaries[0];
	data.inner.boundaries[1] = boundaries[1];
}

BIHNode2::BIHNode2(const BIHNode2& node): index(node.index), data(node.data)
{
}


BIHNode2::BIHNode2()
{
}

BIHNode2::~BIHNode2()
{
}

BIHTree::BIHTree(const Vector& min, const Vector& max): min(min), max(max)
{
}

BIHTree::~BIHTree()
{
	delete this->root;
}

Camera::Camera(Vector position, float hA, float vA): position(position), horizontalAngle(hA), verticalAngle(vA)
{
}

Matrix Matrix::operator*(const float f)
{
	return Matrix(m00*f,m01*f,m02*f,m10*f,m11*f,m12*f,m20*f,m21*f,m22*f);
}
Matrix Matrix::operator*(const float f) const
{
	return Matrix(m00*f,m01*f,m02*f,m10*f,m11*f,m12*f,m20*f,m21*f,m22*f);
}


Matrix Matrix::rotationX(float a)
{
	float sinA = sin(a);
	float cosA = cos(a);
	return Matrix(1,0,0,0,cosA,-sinA,0,sinA,cosA);
}

Matrix Matrix::rotationY(float a)
{
	float sinA = sin(a);
	float cosA = cos(a);
	return Matrix(cosA,0,sinA,0,1,0,-sinA,0,cosA);
}

Matrix Matrix::rotationZ(float a)
{
	float sinA = sin(a);
	float cosA = cos(a);
	return Matrix(cosA,-sinA,0,sinA,cosA,0,0,0,1);
}

Matrix Matrix::rotation(Vector u, float a)
{
	Matrix S(0,u.z,-u.y,-u.z,0,u.x,u.y,-u.x,0);
	return Matrix::UNIT_MATRIX+S*sin(a)+(S*S)*(1-cos(a));
}

Matrix Matrix::operator+(const Matrix& m)
{
	return Matrix(m00+m.m00,m01+m.m01,m02+m.m02,m10+m.m10,m11+m.m11,m12+m.m12,m20+m.m20,m21+m.m21,m22+m.m22);
}

Matrix Matrix::operator+(const Matrix& m) const
{
	return Matrix(m00+m.m00,m01+m.m01,m02+m.m02,m10+m.m10,m11+m.m11,m12+m.m12,m20+m.m20,m21+m.m21,m22+m.m22);
}

Matrix::Matrix(float m00, float m01, float m02, float m10, float m11, float m12, float m20, float m21, float m22): m00(m00), m01(m01), m02(m02), m10(m10), m11(m11), m12(m12), m20(m20), m21(m21), m22(m22)
{
}

Matrix::Matrix(const Matrix& m): m00(m.m00), m01(m.m01), m02(m.m02), m10(m.m10), m11(m.m11), m12(m.m12), m20(m.m20), m21(m.m21), m22(m.m22)
{
}

Matrix::Matrix()
{
}

/*Matrix& Matrix::operator=(const Matrix& m)
{
	
}
*/
const Matrix Matrix::UNIT_MATRIX = Matrix(1,0,0,0,1,0,0,0,1);

Matrix Matrix::operator*(const Matrix& m)
{
	return Matrix(
		m00*m.m00+m01*m.m10+m02*m.m20, m00*m.m01+m01*m.m11+m02*m.m21, m00*m.m02+m01*m.m12+m02*m.m22,
		m10*m.m00+m11*m.m10+m12*m.m20, m10*m.m01+m11*m.m11+m12*m.m21, m10*m.m02+m11*m.m12+m12*m.m22,
		m20*m.m00+m21*m.m10+m22*m.m20, m20*m.m01+m21*m.m11+m22*m.m21, m20*m.m02+m21*m.m12+m22*m.m22);
}


Vector Matrix::operator*(const Vector& v)
{
	return Vector(m00*v.x+m01*v.y+m02*v.z, m10*v.x+m11*v.y+m12*v.z, m20*v.x+m21*v.y+m22*v.z);
}

Matrix Matrix::operator*(const Matrix& m) const
{
	return Matrix(
		m00*m.m00+m01*m.m10+m02*m.m20, m00*m.m01+m01*m.m11+m02*m.m21, m00*m.m02+m01*m.m12+m02*m.m22,
		m10*m.m00+m11*m.m10+m12*m.m20, m10*m.m01+m11*m.m11+m12*m.m21, m10*m.m02+m11*m.m12+m12*m.m22,
		m20*m.m00+m21*m.m10+m22*m.m20, m20*m.m01+m21*m.m11+m22*m.m21, m20*m.m02+m21*m.m12+m22*m.m22);
}

Vector Matrix::operator*(const Vector& v) const
{
	return Vector(m00*v.x+m01*v.y+m02*v.z, m10*v.x+m11*v.y+m12*v.z, m20*v.x+m21*v.y+m22*v.z);
} 

Ray Transform::transform(const Ray & ray) const
{
	return Ray(ray.origin+position, rotationMatrix*ray.direction);
}

Vector Transform::transformPosition(const Vector & v) const
{
	return v+position;
}

Vector Transform::transformDirection(const Vector & v) const
{
	return rotationMatrix*v;
}

Transform::Transform(const Matrix & m, const Matrix & inverse, const Vector & v): rotationMatrix(m), inverseRotationMatrix(inverse), position(v)
{
}

Transform::Transform(const Transform & t): rotationMatrix(t.rotationMatrix), inverseRotationMatrix(t.inverseRotationMatrix), position(t.position)
{
}

Transform::Transform()
{
}

/*Transform& Transform::operator=(const Transform& t)
{
	rotationMatrix = t.rotationMatrix;
	inverseRotationMatrix = t.inverseRotationMatrix;
	position = t.position;
}*/

Vector Transform::transformDirectionInverse(const Vector & v) const
{
	return inverseRotationMatrix * v;
}















