//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#ifndef TYPES_h_
#define TYPES_h_

#include <cstddef>
#include <vector>

using namespace std;

#define MIN_DISTANCE 0.00001f
#define BIAS 0.0001f
#define MAX_DISTANCE 1000000.0f

static const unsigned int modulo[5] = {0, 1, 2, 0, 1};

struct Vector {
	union
	{
		struct
		{
			float x;
			float y;
			float z;
		};
		float data[3];
	};
	static const Vector UNIT_VECTORS[3];
	Vector(const float x, const float y, const float z);
	Vector(const Vector& v);
	Vector();
	//Vector& operator=(const Vector& v);
	Vector(const Vector& v, const int newIndex, const float value);
	float operator*(const Vector& p);
	Vector operator-();
	Vector operator+(const Vector& p);
	Vector operator-(const Vector& p);
	bool operator==(const Vector& p);
	Vector operator*(const float s);
	(float&) operator[](const unsigned int i);
	Vector cross(const Vector& p);
	//float length();
	//float lengthSquared();
	int maxValueAxis() const;
	//Vector normalize();
	Vector mul(const Vector& v);
	Vector mul(const Vector& v) const;
	Vector rotateX(float angle);
	Vector rotateY(float angle);
	Vector rotateZ(float angle);
	// const
	float operator[](const unsigned int i) const;
	float operator*(const Vector& p) const;
	Vector operator+(const Vector& p) const;
	Vector operator-(const Vector& p) const;
	Vector operator-() const;
	Vector operator*(const float s) const;
	Vector normalize() const;
	float length() const;
	float lengthSquared() const;


};

struct Matrix
{
	float m00, m01, m02, m10, m11, m12, m20, m21, m22;
	static Matrix rotationX(float a);
	static Matrix rotationY(float a);
	static Matrix rotationZ(float a);
	static Matrix rotation(Vector around, float a);
	static const Matrix UNIT_MATRIX;
	Matrix operator*(const Matrix& m);
	Matrix operator*(const float f);
	Matrix operator*(const float f) const;
	Vector operator*(const Vector& v);
	Matrix operator*(const Matrix& m) const;
	Vector operator*(const Vector& v) const;
	Matrix operator+(const Matrix& m);
	Matrix operator+(const Matrix& m) const;
	
	Matrix(float m00, float m01, float m02, float m10, float m11, float m12, float m20, float m21, float m22);
	Matrix(const Matrix& m);
	Matrix();
	//Matrix& operator=(const Matrix& m);
};

typedef Vector Normal;
typedef Vector Point;
typedef Vector Color;

struct Ray {
	Point origin;
	Normal direction;
	int sign[3];
	Normal inv_direction;
	Ray(Point o, Normal d);
	Point operator*(float t);
	Point operator*(float t) const;
};

enum LightType
{
	DirectionalLight,
	PointLight,
	AmbientLight
};

struct Light
{
	Vector position;
	Color color;
	LightType type;
	Light(Vector position, LightType type, Color color);
	bool HasDirection();
	/*Color calculateFallin(const Vector& position, const Normal& surfaceNormal, const Color& surfaceColor);
	Ray createShadowRay(const Vector& from);*/
	
};

struct VertexTriangle
{
	Vector a;
	Vector b;
	Vector c;

	float minValue(unsigned int axis);
	float maxValue(unsigned int axis);
	float midValue(unsigned int axis);
};

struct NormalTriangle
{
	Normal a, b, c;
};

struct TriAccel
{
	float n_u;
	float n_v;
	float n_d;
	int k;

	float b_nu;
	float b_nv;
	float b_d;
	int pad;

	float c_nu;
	float c_nv;
	float c_d;
	int pad2;
};

struct BIHNode
{	
	bool isLeaf;
	union {
		struct Leaf
		{
			unsigned int item;
			unsigned int count;
		} leaf;
		struct Inner
		{
			unsigned char axis;
			BIHNode* childs[2];
			float boundaries[2];	
		} inner;
	};
	BIHNode(unsigned int item, unsigned int count);
	BIHNode(unsigned char axis);
	~BIHNode();
};

/*#define AXIS(index) index >> 30
#define INDEX(index) index & 0x3FFFFFFF
#define ISLEAF(index) (index & 0xc0000000) == 0xc0000000*/

#define AXIS(index) index & 3
#define INDEX(index) index >> 2
#define ISLEAF(index) (index & 3) == 3

struct BIHNode2
{

	unsigned int index;
	union {
		struct Leaf
		{
			unsigned int count;
		} leaf;
		struct Inner
		{
			float boundaries[2];	
		} inner;
	} data;
	BIHNode2();
	BIHNode2(const BIHNode2& node);
	BIHNode2(unsigned int startIndex, unsigned int count);
	BIHNode2(unsigned char axis, unsigned int childrenIndex, float boundaries[2]);
	~BIHNode2();
};

class BIHTree
{
	public:
		BIHNode* root;
		const Vector min;
		const Vector max;
		BIHTree(const Vector& min, const Vector& max);
		~BIHTree();
};

class BIHTree2
{
	public:
		const BIHNode2* root;
		const Vector min;
		const Vector max;
		const int depth;
		const int count;
		BIHTree2(const BIHNode2* root, const Vector& min, const Vector& max, const int depth, const int count);
		~BIHTree2();
};

class Scene
{
public:
	Point* points;
	Normal* normals;
	TriAccel* triAccels;
	std::size_t count;
	vector<Light> lights;
	Scene();
	~Scene();
};

struct Transform
{
	Matrix rotationMatrix;
	Matrix inverseRotationMatrix;
	Vector position;
	Ray transform(const Ray & ray) const;
	Vector transformPosition(const Vector & v) const;
	Vector transformDirection(const Vector & v) const;
	Vector transformDirectionInverse(const Vector & v) const;
	Transform(const Matrix & m, const Matrix & inverse, const Vector & v);
	Transform(const Transform& t);
	Transform();
};

struct Camera
{
	Vector position;
	float horizontalAngle;
	float verticalAngle;
	Camera(Vector position, float hA, float vA);
};

#endif /* TYPES_h_ */
